#include "sinussweeptortsa.h"

#include <QJsonArray>
#include <QtMath>

SinusSweepToRTSA::SinusSweepToRTSA(double centerFrequency, double sampleRate, int packetsPerSecond, TxFormat txFormat) :
    mCenterFrequency(centerFrequency),
    mSampleRate(sampleRate),
	mSamplesPerPacket(sampleRate/packetsPerSecond),
	mTxFormat(txFormat),
	mLastPhase(0),
	mCurrentFrequency(0),
    mSamplesArray(nullptr),
	mSamplesArrayInt16(nullptr)
{
    mNetworkAccessManager = new QNetworkAccessManager(this);
    mTimer = new QTimer(this);
	mSampleResendTime = 1000 / (mSampleRate / mSamplesPerPacket) ;

	connect(mTimer, SIGNAL(timeout()), this,SLOT(onGeneratePacket()));
}

SinusSweepToRTSA::~SinusSweepToRTSA()
{
    mTimer->stop();
    mTimer->deleteLater();
    mNetworkAccessManager->deleteLater();
	delete mSamplesArray;
	delete mSamplesArrayInt16;
}

void SinusSweepToRTSA::Start()
{
	qDebug("SinusSweepToRTSA::Start");
	mStreamStartTime = ( QDateTime::currentDateTime().currentMSecsSinceEpoch() );
	mLastPacketEnd = mStreamStartTime/1000.0;
	mSumSamples = 0;
	mCurrentFrequency = 0;

    mTimer->setTimerType(Qt::PreciseTimer);
	mTimer->start(mSampleResendTime); //send period
}

void SinusSweepToRTSA::onGeneratePacket(){

	double sampleTimeUS = 1000000/mSampleRate;

	qint64 deltaT = QDateTime::currentDateTime().currentMSecsSinceEpoch();

#if 0
	mSumSamples += mSamplesPerPacket;
	qDebug() << "Estimated Samples/s" << QString::number((double)mSumSamples/(double)(deltaT - mStreamStartTime) * 1000, 'g', 10);
#endif

	//control loop to regulate samples/s | a negative reawake time will stop the timer if sending and creating the Samples is too slow
	deltaT = deltaT - (mLastPacketEnd*1000);
	double ndiff =  1000*((double)deltaT + 1 ) / sampleTimeUS ;
	double reawaketime = mSampleResendTime -  (sampleTimeUS*(ndiff - mSamplesPerPacket ))/1000 ;
	mTimer->setInterval( reawaketime );

	//get start and stop time for this packet
	double newStart = mLastPacketEnd;
	mLastPacketEnd = newStart + ((mSamplesPerPacket + 1) *  sampleTimeUS)/1000000 ;

	double timeOffset = 0.4; //put it into future

	switch (mTxFormat) {
	case TxFormat::FLOAT32:
		BuildFloat32Samples( newStart + timeOffset , mLastPacketEnd + timeOffset , 1.0/mSampleRate );
		break;
	case TxFormat::INT16:
		BuildInt16Samples( newStart + timeOffset , mLastPacketEnd + timeOffset , 1.0/mSampleRate );
		break;
	case TxFormat::ASCII:
		BuildAsciiSamples( newStart + timeOffset , mLastPacketEnd + timeOffset , 1.0/mSampleRate  );
		break;
	}
}

void SinusSweepToRTSA::BuildFloat32Samples(double startTime, double stopTime, double sampleTime){

	const double TwoPI = M_PI * 2;

	if (mSamplesArray == nullptr)
	{
		mSamplesArray = new float[2*mSamplesPerPacket];
	}

	for (int i = 0; i < mSamplesPerPacket; i++)
	{
		double phase = mLastPhase + TwoPI * mCurrentFrequency * i * sampleTime;
		double I =  qCos( phase );
		double Q =  qSin( phase );

		mCurrentFrequency += 0.1; //frequency step per sample

		if (mCurrentFrequency > mSampleRate) {
			mCurrentFrequency = 0;
		}

		mSamplesArray[2*i] = I*0.1;
		mSamplesArray[2*i + 1] = Q*0.1;
	}

	//save last phase of the sinus
	mLastPhase +=  (TwoPI * mCurrentFrequency *  (mSamplesPerPacket) * sampleTime );
	mLastPhase =   mLastPhase - ((int)(mLastPhase/(TwoPI)) * TwoPI);

	QJsonDocument	jdoc({
		{"startTime", startTime  },
		{"endTime" , stopTime  },
		{"startFrequency", mCenterFrequency -  mSampleRate/2 },
		{"endFrequency" , mCenterFrequency + mSampleRate/2 },
		{"stepFrequency", mSampleRate  },
		{"minPower", -2},
		{"maxPower" , 2},
		{"sampleSize", 2},
		{"sampleDepth" , 1},
		{"payload", "iq"},
		{"format", "float32"},
		{"unit" , "volt"},
		{"samples" , 2*mSamplesPerPacket},
	});

	postDataFloat32(jdoc, mSamplesArray, 2*mSamplesPerPacket);
}

void SinusSweepToRTSA::BuildInt16Samples(double startTime, double stopTime, double sampleTime)
{
	const double TwoPI = M_PI * 2;

	if (mSamplesArrayInt16 == nullptr)
	{
		mSamplesArrayInt16 = new int16_t[2*mSamplesPerPacket];
	}

	for (int i = 0; i < mSamplesPerPacket; i++)
	{
		double phase = mLastPhase + TwoPI * mCurrentFrequency * i * sampleTime;
		double I =  qCos( phase );
		double Q =  qSin( phase );

		mCurrentFrequency += 0.1; //frequency step per sample

		if (mCurrentFrequency > mSampleRate) {
			mCurrentFrequency = 0;
		}

		mSamplesArrayInt16[2*i] = I*32;
		mSamplesArrayInt16[2*i + 1] = Q*32;
	}

	//save last phase of the sinus
	mLastPhase +=  (TwoPI * mCurrentFrequency *  (mSamplesPerPacket) * sampleTime );
	mLastPhase =   mLastPhase - ((int)(mLastPhase/(TwoPI)) * TwoPI);

	QJsonDocument	jdoc({
		{"startTime", startTime },
		{"endTime" , stopTime },
		{"startFrequency", mCenterFrequency -  mSampleRate/2 },
		{"endFrequency" , mCenterFrequency + mSampleRate/2 },
		{"minPower", -2},
		{"maxPower" , 2},
		{"sampleSize", 2},
		{"sampleDepth" , 1},
		{"payload", "iq"},
		{"format", "int16"},
		{"scale", 512.0},
		{"unit" , "volt"},
		{"samples" , 2*mSamplesPerPacket},
	});

	postDataInt16(jdoc, mSamplesArrayInt16, 2*mSamplesPerPacket);
}

void SinusSweepToRTSA::BuildAsciiSamples(double startTime, double stopTime, double sampleTime)
{
	const double TwoPI = M_PI * 2;

	//generate IQ data
	QJsonArray array;
	for(int i = 0; i < mSamplesPerPacket; i++)
	{
		double phase = mLastPhase + TwoPI * mCurrentFrequency * i * sampleTime;
		double I =  qCos( phase );
		double Q =  qSin( phase );

		mCurrentFrequency += 0.1; //frequency step per sample
		if(mCurrentFrequency > mSampleRate)
			mCurrentFrequency = 0;

		//fixed point to minimize the string size in JSON packet
		I = ((int)(I * 100)/100.0) ;
		Q = ((int)(Q * 100)/100.0) ;

		array << I;
		array << Q;
	}

	//save last phase of the sinus
	mLastPhase +=  (TwoPI * mCurrentFrequency *  mSamplesPerPacket * sampleTime );
	mLastPhase =   mLastPhase - ((int)(mLastPhase/(TwoPI)) * TwoPI);

	QJsonDocument	jdoc(    {
								 {"startTime", startTime },
								 {"endTime" , stopTime },
								 {"startFrequency", mCenterFrequency -  mSampleRate/2 },
								 {"endFrequency" , mCenterFrequency + mSampleRate/2 },
								 {"minPower", -2},
								 {"maxPower" , 2},
								 {"sampleSize", 2},
								 {"sampleDepth" , 1},
								 {"payload", "iq"},
								 {"unit" , "volt"},
								 {"samples" , array},
							 });

	postDataAscii(jdoc);
}

void SinusSweepToRTSA::postDataFloat32(QJsonDocument jdoc, float *samplesArray, int nSamples)
{
	QNetworkRequest request(QUrl("http://127.0.0.1:" + QString::number(54664) + "/sample"));
    request.setHeader(QNetworkRequest::ContentTypeHeader, QString("application/json"));
    QByteArray byteArray = jdoc.toJson(QJsonDocument::Compact);
	byteArray.append(0x1e);
	byteArray.append(QByteArray::fromRawData(reinterpret_cast<char *>(samplesArray), nSamples*sizeof(float)));

	QNetworkReply *networkReply = mNetworkAccessManager->post(request, byteArray);

	connect(networkReply, SIGNAL(error(QNetworkReply::NetworkError)), this, SLOT(onError(QNetworkReply::NetworkError)));
	connect(mNetworkAccessManager, SIGNAL(finished(QNetworkReply*)), this, SLOT(onFinished(QNetworkReply*)));
}

void SinusSweepToRTSA::postDataInt16(QJsonDocument jdoc, int16_t *samplesArray, int nSamples)
{
	QNetworkRequest request(QUrl("http://127.0.0.1:" + QString::number(54664) + "/sample"));
    request.setHeader(QNetworkRequest::ContentTypeHeader, QString("application/json"));
    QByteArray byteArray = jdoc.toJson(QJsonDocument::Compact);
    byteArray.append(0x1e);
    byteArray.append(QByteArray::fromRawData(reinterpret_cast<char *>(samplesArray), nSamples*sizeof(int16_t)));
	QNetworkReply *networkReply = mNetworkAccessManager->post(request, byteArray);

	connect(networkReply, SIGNAL(error(QNetworkReply::NetworkError)), this, SLOT(onError(QNetworkReply::NetworkError)));
	connect(mNetworkAccessManager, SIGNAL(finished(QNetworkReply*)), this, SLOT(onFinished(QNetworkReply*)));
}

void SinusSweepToRTSA::postDataAscii(QJsonDocument jdoc)
{
	QNetworkRequest request(QUrl("http://localhost:" + QString::number(54664) + "/sample"));
	request.setHeader(QNetworkRequest::ContentTypeHeader, QString("application/json"));
	QNetworkReply *networkReply = mNetworkAccessManager->post(request, jdoc.toJson(QJsonDocument::Compact));

	connect(networkReply, SIGNAL(error(QNetworkReply::NetworkError)), this, SLOT(onError(QNetworkReply::NetworkError)));
	connect(mNetworkAccessManager, SIGNAL(finished(QNetworkReply*)), this, SLOT(onFinished(QNetworkReply*)));
}

void SinusSweepToRTSA::onError(QNetworkReply::NetworkError)
{
	QNetworkReply* nReply = qobject_cast<QNetworkReply*>( sender() );
	qDebug() << "Network Error: " + nReply->errorString();
	mTimer->stop();
}

void SinusSweepToRTSA::onFinished(QNetworkReply *reply)
{
	reply->deleteLater();
}


